@@ -1,112 +1,117 @@
+from __future__ import annotations
+
+import functools
+import ipaddress
 import os
+import string
 import sys
 import traceback
-from collections import namedtuple
-from collections.abc import Sequence
+import typing as t
+import warnings
+from collections import UserString
+from typing import Collection, Dict, Generic, List, NamedTuple, Tuple, TypeVar, Union, overload
+from urllib.parse import quote as urlquote
 from urllib.parse import urlsplit, urlunsplit
-import ipaddress
+
 import markdown
 
-from mkdocs import utils, theme, plugins
-from mkdocs.config.base import Config, ValidationError
+from mkdocs import plugins, theme, utils
+from mkdocs.config.base import (
+    BaseConfigOption,
+    Config,
+    LegacyConfig,
+    PlainConfigSchemaItem,
+    ValidationError,
+)
+from mkdocs.exceptions import ConfigurationError
+
+T = TypeVar('T')
+SomeConfig = TypeVar('SomeConfig', bound=Config)
 
 
-class BaseConfigOption:
+class SubConfig(Generic[SomeConfig], BaseConfigOption[SomeConfig]):
+    """
+    Subconfig Config Option
 
-    def __init__(self):
-        self.warnings = []
-        self.default = None
+    New: If targeting MkDocs 1.4+, please pass a subclass of Config to the
+    constructor, instead of the old style of a sequence of ConfigOption instances.
+    Validation is then enabled by default.
 
-    def is_required(self):
-        return False
+    A set of `config_options` grouped under a single config option.
+    By default, validation errors and warnings resulting from validating
+    `config_options` are ignored (`validate=False`). Users should typically
+    enable validation with `validate=True`.
+    """
 
-    def validate(self, value):
-        return self.run_validation(value)
+    @overload
+    def __init__(
+        self: SubConfig[SomeConfig], config_class: t.Type[SomeConfig], *, validate: bool = True
+    ):
+        """Create a sub-config in a type-safe way, using fields defined in a Config subclass."""
 
-    def reset_warnings(self):
-        self.warnings = []
+    @overload
+    def __init__(
+        self: SubConfig[LegacyConfig],
+        *config_options: PlainConfigSchemaItem,
+        validate: bool = False,
+    ):
+        """Create an untyped sub-config, using directly passed fields."""
 
-    def pre_validation(self, config, key_name):
-        """
-        Before all options are validated, perform a pre-validation process.
-
-        The pre-validation process method should be implemented by subclasses.
-        """
-
-    def run_validation(self, value):
-        """
-        Perform validation for a value.
-
-        The run_validation method should be implemented by subclasses.
-        """
-        return value
-
-    def post_validation(self, config, key_name):
-        """
-        After all options have passed validation, perform a post-validation
-        process to do any additional changes dependent on other config values.
-
-        The post-validation process method should be implemented by subclasses.
-        """
-
-
-class SubConfig(BaseConfigOption, Config):
-    def __init__(self, *config_options):
-        BaseConfigOption.__init__(self)
-        Config.__init__(self, config_options)
+    def __init__(self, *config_options, validate=None):
+        super().__init__()
         self.default = {}
-
-    def validate(self, value):
-        self.load_dict(value)
-        return self.run_validation(value)
+        if (
+            len(config_options) == 1
+            and isinstance(config_options[0], type)
+            and issubclass(config_options[0], Config)
+        ):
+            if validate is None:
+                validate = True
+            (self._make_config,) = config_options
+        else:
+            self._make_config = functools.partial(LegacyConfig, config_options)
+        self._do_validation = bool(validate)
 
     def run_validation(self, value):
-        Config.validate(self)
-        return self
+        config = self._make_config()
+        try:
+            config.load_dict(value)
+            failed, warnings = config.validate()
+        except ConfigurationError as e:
+            raise ValidationError(str(e))
+
+        if self._do_validation:
+            # Capture errors and warnings
+            self.warnings = warnings
+            if failed:
+                # Get the first failing one
+                key, err = failed[0]
+                raise ValidationError(f"Sub-option {key!r} configuration error: {err}")
+
+        return config
 
 
-class ConfigItems(BaseConfigOption):
+class OptionallyRequired(Generic[T], BaseConfigOption[T]):
     """
-    Config Items Option
+    Soft-deprecated, do not use.
 
-    Validates a list of mappings that all must match the same set of
-    options.
-    """
-    def __init__(self, *config_options, **kwargs):
-        BaseConfigOption.__init__(self)
-        self.item_config = SubConfig(*config_options)
-        self.required = kwargs.get('required', False)
-
-    def __repr__(self):
-        return f'{self.__class__.__name__}: {self.item_config}'
-
-    def run_validation(self, value):
-        if value is None:
-            if self.required:
-                raise ValidationError("Required configuration not provided.")
-            else:
-                return ()
-
-        if not isinstance(value, Sequence):
-            raise ValidationError(f'Expected a sequence of mappings, but a '
-                                  f'{type(value)} was given.')
-
-        return [self.item_config.validate(item) for item in value]
-
-
-class OptionallyRequired(BaseConfigOption):
-    """
     A subclass of BaseConfigOption that adds support for default values and
     required values. It is a base class for config options.
     """
 
-    def __init__(self, default=None, required=False):
+    @overload
+    def __init__(self, default=None):
+        ...
+
+    @overload
+    def __init__(self, default=None, *, required: bool):
+        ...
+
+    def __init__(self, default=None, required=None):
         super().__init__()
         self.default = default
-        self.required = required
-
-    def is_required(self):
-        return self.required
+        self._legacy_required = required
+        self.required = bool(required)
 
     def validate(self, value):
         """
@@ -116,56 +121,136 @@ class OptionallyRequired(BaseConfigOption):
         it is empty but has a default, use that. Finally, call the
         run_validation method on the subclass unless.
         """
-
         if value is None:
             if self.default is not None:
-                if hasattr(self.default, 'copy'):
-                    # ensure no mutable values are assigned
-                    value = self.default.copy()
-                else:
-                    value = self.default
+                value = self.default
             elif not self.required:
-                return
+                return None
             elif self.required:
                 raise ValidationError("Required configuration not provided.")
 
         return self.run_validation(value)
 
 
-class Type(OptionallyRequired):
+class ListOfItems(Generic[T], BaseConfigOption[List[T]]):
+    """
+    Validates a homogeneous list of items.
+
+    E.g. for `config_options.ListOfItems(config_options.Type(int))` a valid item is `[1, 2, 3]`.
+    """
+
+    required: Union[bool, None] = None  # Only for subclasses to set.
+
+    def __init__(self, option_type: BaseConfigOption[T], default=None):
+        super().__init__()
+        self.default = default
+        self.option_type = option_type
+        self.option_type.warnings = self.warnings
+
+    def __repr__(self):
+        return f'{type(self).__name__}: {self.option_type}'
+
+    def pre_validation(self, config, key_name):
+        self._config = config
+        self._key_name = key_name
+
+    def run_validation(self, value):
+        if value is None:
+            if self.required or self.default is None:
+                raise ValidationError("Required configuration not provided.")
+            value = self.default
+        if not isinstance(value, list):
+            raise ValidationError(f'Expected a list of items, but a {type(value)} was given.')
+        if not value:  # Optimization for empty list
+            return value
+
+        fake_config = Config(())
+        try:
+            fake_config.config_file_path = self._config.config_file_path
+        except AttributeError:
+            pass
+
+        # Emulate a config-like environment for pre_validation and post_validation.
+        parent_key_name = getattr(self, '_key_name', '')
+        fake_keys = [f'{parent_key_name}[{i}]' for i in range(len(value))]
+        fake_config.data = dict(zip(fake_keys, value))
+
+        for key_name in fake_config:
+            self.option_type.pre_validation(fake_config, key_name)
+        for key_name in fake_config:
+            # Specifically not running `validate` to avoid the OptionallyRequired effect.
+            fake_config[key_name] = self.option_type.run_validation(fake_config[key_name])
+        for key_name in fake_config:
+            self.option_type.post_validation(fake_config, key_name)
+
+        return [fake_config[k] for k in fake_keys]
+
+
+class ConfigItems(ListOfItems[LegacyConfig]):
+    """
+    Deprecated: Use `ListOfItems(SubConfig(...))` instead of `ConfigItems(...)`.
+
+    Validates a list of mappings that all must match the same set of
+    options.
+    """
+
+    @overload
+    def __init__(self, *config_options: PlainConfigSchemaItem):
+        ...
+
+    @overload
+    def __init__(self, *config_options: PlainConfigSchemaItem, required: bool):
+        ...
+
+    def __init__(self, *config_options: PlainConfigSchemaItem, required=None):
+        super().__init__(SubConfig(*config_options), default=[])
+        self._legacy_required = required
+        self.required = bool(required)
+
+
+class Type(Generic[T], OptionallyRequired[T]):
     """
     Type Config Option
 
     Validate the type of a config option against a given Python type.
     """
 
+    @overload
+    def __init__(self, type_: t.Type[T], length: t.Optional[int] = None, **kwargs):
+        ...
+
+    @overload
+    def __init__(self, type_: Tuple[t.Type[T], ...], length: t.Optional[int] = None, **kwargs):
+        ...
+
     def __init__(self, type_, length=None, **kwargs):
         super().__init__(**kwargs)
         self._type = type_
         self.length = length
 
     def run_validation(self, value):
-
         if not isinstance(value, self._type):
             msg = f"Expected type: {self._type} but received: {type(value)}"
         elif self.length is not None and len(value) != self.length:
-            msg = (f"Expected type: {self._type} with length {self.length}"
-                   f" but received: {value} with length {len(value)}")
+            msg = (
+                f"Expected type: {self._type} with length {self.length}"
+                f" but received: {value!r} with length {len(value)}"
+            )
         else:
             return value
 
         raise ValidationError(msg)
 
 
-class Choice(OptionallyRequired):
+class Choice(Generic[T], OptionallyRequired[T]):
     """
     Choice Config Option
 
     Validate the config option against a strict set of values.
     """
 
-    def __init__(self, choices, **kwargs):
-        super().__init__(**kwargs)
+    def __init__(self, choices: Collection[T], default: t.Optional[T] = None, **kwargs):
+        super().__init__(default=default, **kwargs)
         try:
             length = len(choices)
         except TypeError:
@@ -173,16 +258,15 @@ class Choice(OptionallyRequired):
 
         if not length or isinstance(choices, str):
             raise ValueError(f'Expected iterable of choices, got {choices}')
+        if default is not None and default not in choices:
+            raise ValueError(f'{default!r} is not one of {choices!r}')
 
         self.choices = choices
 
     def run_validation(self, value):
         if value not in self.choices:
-            msg = f"Expected one of: {self.choices} but received: {value}"
-        else:
-            return value
-
-        raise ValidationError(msg)
+            raise ValidationError(f"Expected one of: {self.choices} but received: {value!r}")
+        return value
 
 
 class Deprecated(BaseConfigOption):
@@ -195,7 +279,13 @@ class Deprecated(BaseConfigOption):
     ConfigOption instance, then the value is validated against that type.
     """
 
-    def __init__(self, moved_to=None, message=None, removed=False, option_type=None):
+    def __init__(
+        self,
+        moved_to: t.Optional[str] = None,
+        message: t.Optional[str] = None,
+        removed: bool = False,
+        option_type: t.Optional[BaseConfigOption] = None,
+    ):
         super().__init__()
         self.default = None
         self.moved_to = moved_to
@@ -225,19 +315,17 @@ class Deprecated(BaseConfigOption):
             self.warnings.append(self.message.format(key_name))
 
             if self.moved_to is not None:
-                if '.' not in self.moved_to:
-                    target = config
-                    target_key = self.moved_to
-                else:
-                    move_to, target_key = self.moved_to.rsplit('.', 1)
+                *parent_keys, target_key = self.moved_to.split('.')
+                target = config
 
-                    target = config
-                    for key in move_to.split('.'):
-                        target = target.setdefault(key, {})
+                for key in parent_keys:
+                    if target.get(key) is None:
+                        target[key] = {}
+                    target = target[key]
 
-                        if not isinstance(target, dict):
-                            # We can't move it for the user
-                            return
+                    if not isinstance(target, dict):
+                        # We can't move it for the user
+                        return
 
                 target[target_key] = config.pop(key_name)
 
@@ -252,7 +340,15 @@ class Deprecated(BaseConfigOption):
         self.warnings = self.option.warnings
 
 
-class IpAddress(OptionallyRequired):
+class _IpAddressValue(NamedTuple):
+    host: str
+    port: int
+
+    def __str__(self):
+        return f'{self.host}:{self.port}'
+
+
+class IpAddress(OptionallyRequired[_IpAddressValue]):
     """
     IpAddress Config Option
 
@@ -266,6 +362,8 @@ class IpAddress(OptionallyRequired):
             raise ValidationError("Must be a string of format 'IP:PORT'")
 
         if host != 'localhost':
+            if host.startswith('[') and host.endswith(']'):
+                host = host[1:-1]
             try:
                 # Validate and normalize IP Address
                 host = str(ipaddress.ip_address(host))
@@ -277,33 +375,37 @@ class IpAddress(OptionallyRequired):
         except Exception:
             raise ValidationError(f"'{port}' is not a valid port")
 
-        class Address(namedtuple('Address', 'host port')):
-            def __str__(self):
-                return f'{self.host}:{self.port}'
-
-        return Address(host, port)
+        return _IpAddressValue(host, port)
 
     def post_validation(self, config, key_name):
         host = config[key_name].host
         if key_name == 'dev_addr' and host in ['0.0.0.0', '::']:
             self.warnings.append(
-                (f"The use of the IP address '{host}' suggests a production environment "
-                 "or the use of a proxy to connect to the MkDocs server. However, "
-                 "the MkDocs' server is intended for local development purposes only. "
-                 "Please use a third party production-ready server instead.")
+                f"The use of the IP address '{host}' suggests a production environment "
+                "or the use of a proxy to connect to the MkDocs server. However, "
+                "the MkDocs' server is intended for local development purposes only. "
+                "Please use a third party production-ready server instead."
             )
 
 
-class URL(OptionallyRequired):
+class URL(OptionallyRequired[str]):
     """
     URL Config Option
 
     Validate a URL by requiring a scheme is present.
     """
 
-    def __init__(self, default='', required=False, is_dir=False):
+    @overload
+    def __init__(self, default=None, *, is_dir: bool = False):
+        ...
+
+    @overload
+    def __init__(self, default=None, *, required: bool, is_dir: bool = False):
+        ...
+
+    def __init__(self, default=None, required=None, is_dir: bool = False):
         self.is_dir = is_dir
-        super().__init__(default, required)
+        super().__init__(default, required=required)
 
     def run_validation(self, value):
         if value == '':
@@ -319,17 +421,54 @@ class URL(OptionallyRequired):
                 parsed_url = parsed_url._replace(path=f'{parsed_url.path}/')
             return urlunsplit(parsed_url)
 
-        raise ValidationError(
-            "The URL isn't valid, it should include the http:// (scheme)")
+        raise ValidationError("The URL isn't valid, it should include the http:// (scheme)")
+
+
+class Optional(Generic[T], BaseConfigOption[Union[T, None]]):
+    """Wraps a field and makes a None value possible for it when no value is set.
+
+    E.g. `my_field = config_options.Optional(config_options.Type(str))`
+    """
+
+    def __init__(self, config_option: BaseConfigOption[T]):
+        if config_option.default is not None:
+            raise ValueError(
+                f"This option already has a default ({config_option.default!r}) "
+                f"and doesn't need to be wrapped into Optional"
+            )
+        super().__init__()
+        self.option = config_option
+        self.warnings = config_option.warnings
+
+    def __getattr__(self, key):
+        if key in ('option', 'warnings'):
+            raise AttributeError
+        return getattr(self.option, key)
+
+    def pre_validation(self, config, key_name):
+        return self.option.pre_validation(config, key_name)
+
+    def run_validation(self, value):
+        if value is None:
+            return None
+        return self.option.validate(value)
+
+    def post_validation(self, config, key_name):
+        result = self.option.post_validation(config, key_name)
+        self.warnings = self.option.warnings
+        return result
+
+    def reset_warnings(self):
+        self.option.reset_warnings()
+        self.warnings = self.option.warnings
 
 
 class RepoURL(URL):
-    """
-    Repo URL Config Option
-
-    A small extension to the URL config that sets the repo_name and edit_uri,
-    based on the url if they haven't already been provided.
-    """
+    def __init__(self, *args, **kwargs):
+        warnings.warn(
+            "RepoURL is no longer used in MkDocs and will be removed.", DeprecationWarning
+        )
+        super().__init__(*args, **kwargs)
 
     def post_validation(self, config, key_name):
         repo_host = urlsplit(config['repo_url']).netloc.lower()
@@ -362,25 +501,112 @@ class RepoURL(URL):
         config['edit_uri'] = edit_uri
 
 
-class FilesystemObject(Type):
+class EditURI(Type[str]):
+    def __init__(self, repo_url_key: str):
+        super().__init__(str)
+        self.repo_url_key = repo_url_key
+
+    def post_validation(self, config, key_name):
+        edit_uri = config.get(key_name)
+        repo_url = config.get(self.repo_url_key)
+
+        if edit_uri is None and repo_url is not None:
+            repo_host = urlsplit(repo_url).netloc.lower()
+            if repo_host == 'github.com' or repo_host == 'gitlab.com':
+                edit_uri = 'edit/master/docs/'
+            elif repo_host == 'bitbucket.org':
+                edit_uri = 'src/default/docs/'
+
+        # ensure a well-formed edit_uri
+        if edit_uri and not edit_uri.endswith('/'):
+            edit_uri += '/'
+
+        config[key_name] = edit_uri
+
+
+class EditURITemplate(BaseConfigOption[str]):
+    class Formatter(string.Formatter):
+        def convert_field(self, value, conversion):
+            if conversion == 'q':
+                return urlquote(value, safe='')
+            return super().convert_field(value, conversion)
+
+    class Template(UserString):
+        def __init__(self, formatter, data):
+            super().__init__(data)
+            self.formatter = formatter
+            try:
+                self.format('', '')
+            except KeyError as e:
+                raise ValueError(f"Unknown template substitute: {e}")
+
+        def format(self, path, path_noext):
+            return self.formatter.format(self.data, path=path, path_noext=path_noext)
+
+    def __init__(self, edit_uri_key=None):
+        super().__init__()
+        self.edit_uri_key = edit_uri_key
+
+    def run_validation(self, value):
+        try:
+            return self.Template(self.Formatter(), value)
+        except Exception as e:
+            raise ValidationError(e)
+
+    def post_validation(self, config, key_name):
+        if self.edit_uri_key and config.get(key_name) and config.get(self.edit_uri_key):
+            self.warnings.append(
+                f"The option '{self.edit_uri_key}' has no effect when '{key_name}' is set."
+            )
+
+
+class RepoName(Type[str]):
+    def __init__(self, repo_url_key: str):
+        super().__init__(str)
+        self.repo_url_key = repo_url_key
+
+    def post_validation(self, config, key_name):
+        repo_name = config.get(key_name)
+        repo_url = config.get(self.repo_url_key)
+
+        # derive repo_name from repo_url if unset
+        if repo_url is not None and repo_name is None:
+            repo_host = urlsplit(config['repo_url']).netloc.lower()
+            if repo_host == 'github.com':
+                repo_name = 'GitHub'
+            elif repo_host == 'bitbucket.org':
+                repo_name = 'Bitbucket'
+            elif repo_host == 'gitlab.com':
+                repo_name = 'GitLab'
+            else:
+                repo_name = repo_host.split('.')[0].title()
+            config[key_name] = repo_name
+
+
+class FilesystemObject(Type[str]):
     """
     Base class for options that point to filesystem objects.
     """
 
-    def __init__(self, exists=False, **kwargs):
+    existence_test = staticmethod(os.path.exists)
+    name = 'file or directory'
+
+    def __init__(self, exists: bool = False, **kwargs):
         super().__init__(type_=str, **kwargs)
         self.exists = exists
         self.config_dir = None
 
     def pre_validation(self, config, key_name):
-        self.config_dir = os.path.dirname(config.config_file_path) if config.config_file_path else None
+        self.config_dir = (
+            os.path.dirname(config.config_file_path) if config.config_file_path else None
+        )
 
     def run_validation(self, value):
         value = super().run_validation(value)
         if self.config_dir and not os.path.isabs(value):
             value = os.path.join(self.config_dir, value)
         if self.exists and not self.existence_test(value):
-            raise ValidationError(f"The path {value} isn't an existing {self.name}.")
+            raise ValidationError(f"The path '{value}' isn't an existing {self.name}.")
         return os.path.abspath(value)
 
 
@@ -390,9 +616,12 @@ class Dir(FilesystemObject):
 
     Validate a path to a directory, optionally verifying that it exists.
     """
+
     existence_test = staticmethod(os.path.isdir)
     name = 'directory'
 
+
+class DocsDir(Dir):
     def post_validation(self, config, key_name):
         if config.config_file_path is None:
             return
@@ -400,9 +629,10 @@ class Dir(FilesystemObject):
         # Validate that the dir is not the parent dir of the config file.
         if os.path.dirname(config.config_file_path) == config[key_name]:
             raise ValidationError(
-                (f"The '{key_name}' should not be the parent directory of the"
-                 " config file. Use a child directory instead so that the"
-                 f" '{key_name}' is a sibling of the config file."))
+                f"The '{key_name}' should not be the parent directory of the"
+                f" config file. Use a child directory instead so that the"
+                f" '{key_name}' is a sibling of the config file."
+            )
 
 
 class File(FilesystemObject):
@@ -411,38 +641,33 @@ class File(FilesystemObject):
 
     Validate a path to a file, optionally verifying that it exists.
     """
+
     existence_test = staticmethod(os.path.isfile)
     name = 'file'
 
 
-class ListOfPaths(OptionallyRequired):
+class ListOfPaths(ListOfItems[str]):
     """
     List of Paths Config Option
 
     A list of file system paths. Raises an error if one of the paths does not exist.
+
+    For greater flexibility, prefer ListOfItems, e.g. to require files specifically:
+
+        config_options.ListOfItems(config_options.File(exists=True))
     """
 
-    def __init__(self, default=[], required=False):
-        self.config_dir = None
-        super().__init__(default, required)
+    @overload
+    def __init__(self, default=[]):
+        ...
 
-    def pre_validation(self, config, key_name):
-        self.config_dir = os.path.dirname(config.config_file_path) if config.config_file_path else None
+    @overload
+    def __init__(self, default=[], *, required: bool):
+        ...
 
-    def run_validation(self, value):
-        if not isinstance(value, list):
-            raise ValidationError(f"Expected a list, got {type(value)}")
-        if len(value) == 0:
-            return
-        paths = []
-        for path in value:
-            if self.config_dir and not os.path.isabs(path):
-                path = os.path.join(self.config_dir, path)
-            if not os.path.exists(path):
-                raise ValidationError(f"The path {path} does not exist.")
-            path = os.path.abspath(path)
-            paths.append(path)
-        return paths
+    def __init__(self, default=[], required=None):
+        super().__init__(FilesystemObject(exists=True), default)
+        self.required = required
 
 
 class SiteDir(Dir):
@@ -453,29 +678,30 @@ class SiteDir(Dir):
     """
 
     def post_validation(self, config, key_name):
-
         super().post_validation(config, key_name)
+        docs_dir = config['docs_dir']
+        site_dir = config['site_dir']
 
         # Validate that the docs_dir and site_dir don't contain the
         # other as this will lead to copying back and forth on each
         # and eventually make a deep nested mess.
-        if (config['docs_dir'] + os.sep).startswith(config['site_dir'].rstrip(os.sep) + os.sep):
+        if (docs_dir + os.sep).startswith(site_dir.rstrip(os.sep) + os.sep):
             raise ValidationError(
-                ("The 'docs_dir' should not be within the 'site_dir' as this "
-                 "can mean the source files are overwritten by the output or "
-                 "it will be deleted if --clean is passed to mkdocs build."
-                 "(site_dir: '{}', docs_dir: '{}')"
-                 ).format(config['site_dir'], config['docs_dir']))
-        elif (config['site_dir'] + os.sep).startswith(config['docs_dir'].rstrip(os.sep) + os.sep):
+                f"The 'docs_dir' should not be within the 'site_dir' as this "
+                f"can mean the source files are overwritten by the output or "
+                f"it will be deleted if --clean is passed to mkdocs build. "
+                f"(site_dir: '{site_dir}', docs_dir: '{docs_dir}')"
+            )
+        elif (site_dir + os.sep).startswith(docs_dir.rstrip(os.sep) + os.sep):
             raise ValidationError(
-                ("The 'site_dir' should not be within the 'docs_dir' as this "
-                 "leads to the build directory being copied into itself and "
-                 "duplicate nested files in the 'site_dir'."
-                 "(site_dir: '{}', docs_dir: '{}')"
-                 ).format(config['site_dir'], config['docs_dir']))
+                f"The 'site_dir' should not be within the 'docs_dir' as this "
+                f"leads to the build directory being copied into itself and "
+                f"duplicate nested files in the 'site_dir'. "
+                f"(site_dir: '{site_dir}', docs_dir: '{docs_dir}')"
+            )
 
 
-class Theme(BaseConfigOption):
+class Theme(BaseConfigOption[theme.Theme]):
     """
     Theme Config Option
 
@@ -486,7 +712,7 @@ class Theme(BaseConfigOption):
         super().__init__()
         self.default = default
 
-    def validate(self, value):
+    def run_validation(self, value):
         if value is None and self.default is not None:
             value = {'name': self.default}
 
@@ -507,13 +733,15 @@ class Theme(BaseConfigOption):
 
             raise ValidationError("No theme name set.")
 
-        raise ValidationError(f'Invalid type "{type(value)}". Expected a string or key/value pairs.')
+        raise ValidationError(f'Invalid type {type(value)}. Expected a string or key/value pairs.')
 
     def post_validation(self, config, key_name):
         theme_config = config[key_name]
 
         if not theme_config['name'] and 'custom_dir' not in theme_config:
-            raise ValidationError("At least one of 'theme.name' or 'theme.custom_dir' must be defined.")
+            raise ValidationError(
+                f"At least one of '{key_name}.name' or '{key_name}.custom_dir' must be defined."
+            )
 
         # Ensure custom_dir is an absolute path
         if 'custom_dir' in theme_config and not os.path.isabs(theme_config['custom_dir']):
@@ -521,11 +749,14 @@ class Theme(BaseConfigOption):
             theme_config['custom_dir'] = os.path.join(config_dir, theme_config['custom_dir'])
 
         if 'custom_dir' in theme_config and not os.path.isdir(theme_config['custom_dir']):
-            raise ValidationError("The path set in {name}.custom_dir ('{path}') does not exist.".
-                                  format(path=theme_config['custom_dir'], name=key_name))
+            raise ValidationError(
+                "The path set in {name}.custom_dir ('{path}') does not exist.".format(
+                    path=theme_config['custom_dir'], name=key_name
+                )
+            )
 
         if 'locale' in theme_config and not isinstance(theme_config['locale'], str):
-            raise ValidationError(f"'{theme_config['name']}.locale' must be a string.")
+            raise ValidationError(f"'{key_name}.locale' must be a string.")
 
         config[key_name] = theme.Theme(**theme_config)
 
@@ -559,11 +790,15 @@ class Nav(OptionallyRequired):
             pass
         elif isinstance(value, dict):
             if len(value) != 1:
-                raise ValidationError(f"Expected nav item to be a dict of size 1, got {self._repr_item(value)}")
+                raise ValidationError(
+                    f"Expected nav item to be a dict of size 1, got {self._repr_item(value)}"
+                )
             for subnav in value.values():
                 self.run_validation(subnav, top=False)
         else:
-            raise ValidationError(f"Expected nav item to be a string or dict, got {self._repr_item(value)}")
+            raise ValidationError(
+                f"Expected nav item to be a string or dict, got {self._repr_item(value)}"
+            )
 
     @classmethod
     def _repr_item(cls, value):
@@ -575,7 +810,7 @@ class Nav(OptionallyRequired):
             return f"a {type(value).__name__}: {value!r}"
 
 
-class Private(OptionallyRequired):
+class Private(BaseConfigOption):
     """
     Private Config Option
 
@@ -583,10 +818,11 @@ class Private(OptionallyRequired):
     """
 
     def run_validation(self, value):
-        raise ValidationError('For internal use only.')
+        if value is not None:
+            raise ValidationError('For internal use only.')
 
 
-class MarkdownExtensions(OptionallyRequired):
+class MarkdownExtensions(OptionallyRequired[List[str]]):
     """
     Markdown Extensions Config Option
 
@@ -595,12 +831,18 @@ class MarkdownExtensions(OptionallyRequired):
     a dict item must be a valid Markdown extension name and the value must be a dict of config
     options for that extension. Extension configs are set on the private setting passed to
     `configkey`. The `builtins` keyword accepts a list of extensions which cannot be overridden by
-    the user. However, builtins can be duplicated to define config options for them if desired. """
-    def __init__(self, builtins=None, configkey='mdx_configs', **kwargs):
-        super().__init__(**kwargs)
+    the user. However, builtins can be duplicated to define config options for them if desired."""
+
+    def __init__(
+        self,
+        builtins: t.Optional[List[str]] = None,
+        configkey: str = 'mdx_configs',
+        default: List[str] = [],
+        **kwargs,
+    ):
+        super().__init__(default=default, **kwargs)
         self.builtins = builtins or []
         self.configkey = configkey
-        self.configdata = {}
 
     def validate_ext_cfg(self, ext, cfg):
         if not isinstance(ext, str):
@@ -612,6 +854,7 @@ class MarkdownExtensions(OptionallyRequired):
         self.configdata[ext] = cfg
 
     def run_validation(self, value):
+        self.configdata = {}
         if not isinstance(value, (list, tuple, dict)):
             raise ValidationError('Invalid Markdown Extensions configuration')
         extensions = []
@@ -647,7 +890,9 @@ class MarkdownExtensions(OptionallyRequired):
                     stack.insert(0, frame)
                 tb = ''.join(traceback.format_list(stack))
 
-                raise ValidationError(f"Failed to load extension '{ext}'.\n{tb}{type(e).__name__}: {e}")
+                raise ValidationError(
+                    f"Failed to load extension '{ext}'.\n{tb}{type(e).__name__}: {e}"
+                )
 
         return extensions
 
@@ -655,7 +900,7 @@ class MarkdownExtensions(OptionallyRequired):
         config[self.configkey] = self.configdata
 
 
-class Plugins(OptionallyRequired):
+class Plugins(OptionallyRequired[plugins.PluginCollection]):
     """
     Plugins config option.
 
@@ -667,6 +912,7 @@ class Plugins(OptionallyRequired):
         super().__init__(**kwargs)
         self.installed_plugins = plugins.get_plugins()
         self.config_file_path = None
+        self.plugin_cache: Dict[str, plugins.BasePlugin] = {}
 
     def pre_validation(self, config, key_name):
         self.config_file_path = config.config_file_path
@@ -674,10 +920,10 @@ class Plugins(OptionallyRequired):
     def run_validation(self, value):
         if not isinstance(value, (list, tuple, dict)):
             raise ValidationError('Invalid Plugins configuration. Expected a list or dict.')
-        plgins = plugins.PluginCollection()
+        self.plugins = plugins.PluginCollection()
         if isinstance(value, dict):
             for name, cfg in value.items():
-                plgins[name] = self.load_plugin(name, cfg)
+                self.plugins[name] = self.load_plugin(name, cfg)
         else:
             for item in value:
                 if isinstance(item, dict):
@@ -687,8 +933,8 @@ class Plugins(OptionallyRequired):
                     item = name
                 else:
                     cfg = {}
-                plgins[item] = self.load_plugin(item, cfg)
-        return plgins
+                self.plugins[item] = self.load_plugin(item, cfg)
+        return self.plugins
 
     def load_plugin(self, name, config):
         if not isinstance(name, str):
@@ -700,20 +946,56 @@ class Plugins(OptionallyRequired):
         if not isinstance(config, dict):
             raise ValidationError(f"Invalid config options for the '{name}' plugin.")
 
-        Plugin = self.installed_plugins[name].load()
+        try:
+            plugin = self.plugin_cache[name]
+        except KeyError:
+            Plugin = self.installed_plugins[name].load()
 
-        if not issubclass(Plugin, plugins.BasePlugin):
-            raise ValidationError(
-                f'{Plugin.__module__}.{Plugin.__name__} must be a subclass of'
-                f' {plugins.BasePlugin.__module__}.{plugins.BasePlugin.__name__}')
+            if not issubclass(Plugin, plugins.BasePlugin):
+                raise ValidationError(
+                    f'{Plugin.__module__}.{Plugin.__name__} must be a subclass of'
+                    f' {plugins.BasePlugin.__module__}.{plugins.BasePlugin.__name__}'
+                )
+
+            plugin = Plugin()
+
+            if hasattr(plugin, 'on_startup') or hasattr(plugin, 'on_shutdown'):
+                self.plugin_cache[name] = plugin
 
-        plugin = Plugin()
         errors, warnings = plugin.load_config(config, self.config_file_path)
         self.warnings.extend(warnings)
-        errors_message = '\n'.join(
-            f"Plugin '{name}' value: '{x}'. Error: {y}"
-            for x, y in errors
-        )
+        errors_message = '\n'.join(f"Plugin '{name}' value: '{x}'. Error: {y}" for x, y in errors)
         if errors_message:
             raise ValidationError(errors_message)
         return plugin
+
+
+class Hooks(ListOfItems):
+    """A list of Python scripts to be treated as instances of plugins."""
+
+    def __init__(self, plugins_key: str):
+        super().__init__(File(exists=True), default=[])
+        self.plugins_key = plugins_key
+
+    def run_validation(self, value):
+        paths = super().run_validation(value)
+        hooks = {}
+        for name, path in zip(value, paths):
+            hooks[name] = self._load_hook(name, path)
+        return hooks
+
+    @functools.lru_cache(maxsize=None)
+    def _load_hook(self, name, path):
+        import importlib.util
+
+        spec = importlib.util.spec_from_file_location(name, path)
+        if spec is None:
+            raise ValidationError(f"Cannot import path '{path}' as a Python module")
+        module = importlib.util.module_from_spec(spec)
+        spec.loader.exec_module(module)
+        return module
+
+    def post_validation(self, config, key_name):
+        plugins = config[self.plugins_key]
+        for name, hook in config[key_name].items():
+            plugins[name] = hook
