@@ -1,17 +1,20 @@
-import unittest
+import functools
 import os
 import sys
+import unittest
 from unittest import mock
-from tempfile import TemporaryDirectory
 
-from mkdocs.structure.pages import Page
 from mkdocs.structure.files import File, Files
-from mkdocs.tests.base import load_config, dedent
+from mkdocs.structure.pages import Page
+from mkdocs.tests.base import dedent, load_config, tempdir
+
+load_config = functools.lru_cache(maxsize=None)(load_config)
 
 
 class PageTests(unittest.TestCase):
-
-    DOCS_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../integration/subpages/docs')
+    DOCS_DIR = os.path.join(
+        os.path.abspath(os.path.dirname(__file__)), '../integration/subpages/docs'
+    )
 
     def test_homepage(self):
         cfg = load_config(docs_dir=self.DOCS_DIR)
@@ -410,239 +413,175 @@ class PageTests(unittest.TestCase):
         # Different File
         self.assertTrue(pg != Page('Foo', f2, cfg))
 
-    def test_BOM(self):
+    @tempdir()
+    def test_BOM(self, docs_dir):
         md_src = '# An UTF-8 encoded file with a BOM'
-        with TemporaryDirectory() as docs_dir:
-            # We don't use mkdocs.tests.base.tempdir decorator here due to uniqueness of this test.
-            cfg = load_config(docs_dir=docs_dir)
-            fl = File('index.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
-            pg = Page(None, fl, cfg)
-            # Create an UTF-8 Encoded file with BOM (as Microsoft editors do). See #1186
-            with open(fl.abs_src_path, 'w', encoding='utf-8-sig') as f:
-                f.write(md_src)
-            # Now read the file.
-            pg.read_source(cfg)
-            # Ensure the BOM (`\ufeff`) is removed
-            self.assertNotIn('\ufeff', pg.markdown)
-            self.assertEqual(pg.markdown, md_src)
-            self.assertEqual(pg.meta, {})
+        cfg = load_config(docs_dir=docs_dir)
+        fl = File('index.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
+        pg = Page(None, fl, cfg)
+        # Create an UTF-8 Encoded file with BOM (as Microsoft editors do). See #1186
+        with open(fl.abs_src_path, 'w', encoding='utf-8-sig') as f:
+            f.write(md_src)
+        # Now read the file.
+        pg.read_source(cfg)
+        # Ensure the BOM (`\ufeff`) is removed
+        self.assertNotIn('\ufeff', pg.markdown)
+        self.assertEqual(pg.markdown, md_src)
+        self.assertEqual(pg.meta, {})
 
-    def test_page_edit_url(self):
-        configs = [
-            {
-                'repo_url': 'http://github.com/mkdocs/mkdocs'
-            },
-            {
-                'repo_url': 'https://github.com/mkdocs/mkdocs/'
-            }, {
-                'repo_url': 'http://example.com'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': 'edit/master'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '?query=edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '?query=edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '#edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '#edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': ''  # Set to blank value
-            }, {
-                # Nothing defined
-            }
-        ]
-
-        expected = [
-            'http://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
-            'https://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
-            None,
-            'http://example.com/edit/master/testing.md',
-            'http://example.com/edit/master/testing.md',
-            'http://example.com/edit/master/testing.md',
-            'http://example.com/edit/master/testing.md',
-            'http://example.com/edit/master/testing.md',
-            'http://example.com/foo/edit/master/testing.md',
-            'http://example.com/foo/edit/master/testing.md',
-            'http://example.com?query=edit/master/testing.md',
-            'http://example.com/?query=edit/master/testing.md',
-            'http://example.com#edit/master/testing.md',
-            'http://example.com/#edit/master/testing.md',
-            None,
-            None
-        ]
-
-        for i, c in enumerate(configs):
-            cfg = load_config(**c)
-            fl = File('testing.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
-            pg = Page('Foo', fl, cfg)
-            self.assertEqual(pg.url, 'testing/')
-            self.assertEqual(pg.edit_url, expected[i])
-
-    def test_nested_page_edit_url(self):
-        configs = [
-            {
-                'repo_url': 'http://github.com/mkdocs/mkdocs'
-            },
-            {
-                'repo_url': 'https://github.com/mkdocs/mkdocs/'
-            }, {
-                'repo_url': 'http://example.com'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': 'edit/master'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '?query=edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '?query=edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '#edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '#edit/master/'
-            }
-        ]
-
-        expected = [
-            'http://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
-            'https://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
-            None,
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/foo/edit/master/sub1/non-index.md',
-            'http://example.com/foo/edit/master/sub1/non-index.md',
-            'http://example.com?query=edit/master/sub1/non-index.md',
-            'http://example.com/?query=edit/master/sub1/non-index.md',
-            'http://example.com#edit/master/sub1/non-index.md',
-            'http://example.com/#edit/master/sub1/non-index.md'
-        ]
-
-        for i, c in enumerate(configs):
-            c['docs_dir'] = self.DOCS_DIR
-            cfg = load_config(**c)
-            fl = File('sub1/non-index.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
-            pg = Page('Foo', fl, cfg)
-            self.assertEqual(pg.url, 'sub1/non-index/')
-            self.assertEqual(pg.edit_url, expected[i])
+    def test_page_edit_url(
+        self, paths={'testing.md': 'testing/', 'sub1/non-index.md': 'sub1/non-index/'}
+    ):
+        for case in [
+            dict(
+                config={'repo_url': 'http://github.com/mkdocs/mkdocs'},
+                edit_url='http://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
+                edit_url2='http://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'https://github.com/mkdocs/mkdocs/'},
+                edit_url='https://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
+                edit_url2='https://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com'},
+                edit_url=None,
+                edit_url2=None,
+            ),
+            dict(
+                config={'repo_url': 'http://example.com', 'edit_uri': 'edit/master'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com', 'edit_uri': '/edit/master'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/foo/', 'edit_uri': '/edit/master/'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/foo', 'edit_uri': '/edit/master/'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/foo/', 'edit_uri': '/edit/master'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/foo/', 'edit_uri': 'edit/master/'},
+                edit_url='http://example.com/foo/edit/master/testing.md',
+                edit_url2='http://example.com/foo/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/foo', 'edit_uri': 'edit/master/'},
+                edit_url='http://example.com/foo/edit/master/testing.md',
+                edit_url2='http://example.com/foo/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com', 'edit_uri': '?query=edit/master'},
+                edit_url='http://example.com?query=edit/master/testing.md',
+                edit_url2='http://example.com?query=edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/', 'edit_uri': '?query=edit/master/'},
+                edit_url='http://example.com/?query=edit/master/testing.md',
+                edit_url2='http://example.com/?query=edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com', 'edit_uri': '#edit/master'},
+                edit_url='http://example.com#edit/master/testing.md',
+                edit_url2='http://example.com#edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'repo_url': 'http://example.com/', 'edit_uri': '#edit/master/'},
+                edit_url='http://example.com/#edit/master/testing.md',
+                edit_url2='http://example.com/#edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'edit_uri': 'http://example.com/edit/master'},
+                edit_url='http://example.com/edit/master/testing.md',
+                edit_url2='http://example.com/edit/master/sub1/non-index.md',
+            ),
+            dict(
+                config={'edit_uri_template': 'https://github.com/project/repo/wiki/{path_noext}'},
+                edit_url='https://github.com/project/repo/wiki/testing',
+                edit_url2='https://github.com/project/repo/wiki/sub1/non-index',
+            ),
+            dict(
+                config={
+                    'repo_url': 'https://github.com/project/repo/wiki',
+                    'edit_uri_template': '{path_noext}/_edit',
+                },
+                edit_url='https://github.com/project/repo/wiki/testing/_edit',
+                edit_url2='https://github.com/project/repo/wiki/sub1/non-index/_edit',
+            ),
+            dict(
+                config={
+                    'repo_url': 'https://gitlab.com/project/repo',
+                    'edit_uri_template': '-/sse/master/docs%2F{path!q}',
+                },
+                edit_url='https://gitlab.com/project/repo/-/sse/master/docs%2Ftesting.md',
+                edit_url2='https://gitlab.com/project/repo/-/sse/master/docs%2Fsub1%2Fnon-index.md',
+            ),
+            dict(
+                config={
+                    'repo_url': 'https://bitbucket.org/project/repo/',
+                    'edit_uri_template': 'src/master/docs/{path}?mode=edit',
+                },
+                edit_url='https://bitbucket.org/project/repo/src/master/docs/testing.md?mode=edit',
+                edit_url2='https://bitbucket.org/project/repo/src/master/docs/sub1/non-index.md?mode=edit',
+            ),
+            dict(
+                config={
+                    'repo_url': 'http://example.com',
+                    'edit_uri': '',
+                    'edit_uri_template': '',
+                },  # Set to blank value
+                edit_url=None,
+                edit_url2=None,
+            ),
+            dict(config={}, edit_url=None, edit_url2=None),  # Nothing defined
+        ]:
+            for i, path in enumerate(paths, 1):
+                edit_url_key = f'edit_url{i}' if i > 1 else 'edit_url'
+                with self.subTest(case['config'], path=path):
+                    cfg = load_config(**case['config'])
+                    fl = File(path, cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
+                    pg = Page('Foo', fl, cfg)
+                    self.assertEqual(pg.url, paths[path])
+                    self.assertEqual(pg.edit_url, case[edit_url_key])
 
     @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
-    def test_nested_page_edit_url_windows(self):
-        configs = [
-            {
-                'repo_url': 'http://github.com/mkdocs/mkdocs'
-            },
-            {
-                'repo_url': 'https://github.com/mkdocs/mkdocs/'
-            }, {
-                'repo_url': 'http://example.com'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': 'edit/master'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': '/edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': '/edit/master'
-            }, {
-                'repo_url': 'http://example.com/foo/',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com/foo',
-                'edit_uri': 'edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '?query=edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '?query=edit/master/'
-            }, {
-                'repo_url': 'http://example.com',
-                'edit_uri': '#edit/master'
-            }, {
-                'repo_url': 'http://example.com/',
-                'edit_uri': '#edit/master/'
-            }
-        ]
+    def test_page_edit_url_windows(self):
+        self.test_page_edit_url(
+            paths={'testing.md': 'testing/', 'sub1\\non-index.md': 'sub1/non-index/'}
+        )
 
-        expected = [
-            'http://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
-            'https://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
-            None,
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/edit/master/sub1/non-index.md',
-            'http://example.com/foo/edit/master/sub1/non-index.md',
-            'http://example.com/foo/edit/master/sub1/non-index.md',
-            'http://example.com?query=edit/master/sub1/non-index.md',
-            'http://example.com/?query=edit/master/sub1/non-index.md',
-            'http://example.com#edit/master/sub1/non-index.md',
-            'http://example.com/#edit/master/sub1/non-index.md'
-        ]
-
-        for i, c in enumerate(configs):
-            c['docs_dir'] = self.DOCS_DIR
-            cfg = load_config(**c)
-            fl = File('sub1\\non-index.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
-            pg = Page('Foo', fl, cfg)
-            self.assertEqual(pg.url, 'sub1/non-index/')
-            self.assertEqual(pg.edit_url, expected[i])
+    def test_page_edit_url_warning(self):
+        for case in [
+            dict(
+                config={'edit_uri': 'edit/master'},
+                edit_url='edit/master/testing.md',
+                warning="WARNING:mkdocs.structure.pages:edit_uri: "
+                "'edit/master/testing.md' is not a valid URL, it should include the http:// (scheme)",
+            ),
+        ]:
+            with self.subTest(case['config']):
+                with self.assertLogs('mkdocs') as cm:
+                    cfg = load_config(**case['config'])
+                    fl = File(
+                        'testing.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls']
+                    )
+                    pg = Page('Foo', fl, cfg)
+                self.assertEqual(pg.url, 'testing/')
+                self.assertEqual(pg.edit_url, case['edit_url'])
+                self.assertEqual(cm.output, [case['warning']])
 
     def test_page_render(self):
         cfg = load_config()
@@ -652,25 +591,33 @@ class PageTests(unittest.TestCase):
         self.assertEqual(pg.content, None)
         self.assertEqual(pg.toc, [])
         pg.render(cfg, [fl])
-        self.assertTrue(pg.content.startswith(
-            '<h1 id="welcome-to-mkdocs">Welcome to MkDocs</h1>\n'
-        ))
-        self.assertEqual(str(pg.toc).strip(), dedent("""
-            Welcome to MkDocs - #welcome-to-mkdocs
-                Commands - #commands
-                Project layout - #project-layout
-        """))
+        self.assertTrue(
+            pg.content.startswith('<h1 id="welcome-to-mkdocs">Welcome to MkDocs</h1>\n')
+        )
+        self.assertEqual(
+            str(pg.toc).strip(),
+            dedent(
+                """
+                Welcome to MkDocs - #welcome-to-mkdocs
+                    Commands - #commands
+                    Project layout - #project-layout
+                """
+            ),
+        )
 
     def test_missing_page(self):
         cfg = load_config()
         fl = File('missing.md', cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls'])
         pg = Page('Foo', fl, cfg)
-        with self.assertRaises(OSError):
-            pg.read_source(cfg)
+        with self.assertLogs('mkdocs') as cm:
+            with self.assertRaises(OSError):
+                pg.read_source(cfg)
+        self.assertEqual(
+            '\n'.join(cm.output), 'ERROR:mkdocs.structure.pages:File not found: missing.md'
+        )
 
 
 class SourceDateEpochTests(unittest.TestCase):
-
     def setUp(self):
         self.default = os.environ.get('SOURCE_DATE_EPOCH', None)
         os.environ['SOURCE_DATE_EPOCH'] = '0'
@@ -689,14 +636,13 @@ class SourceDateEpochTests(unittest.TestCase):
 
 
 class RelativePathExtensionTests(unittest.TestCase):
-
-    DOCS_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../integration/subpages/docs')
+    DOCS_DIR = os.path.join(
+        os.path.abspath(os.path.dirname(__file__)), '../integration/subpages/docs'
+    )
 
     def get_rendered_result(self, files):
         cfg = load_config(docs_dir=self.DOCS_DIR)
-        fs = []
-        for f in files:
-            fs.append(File(f.replace('/', os.sep), cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls']))
+        fs = [File(f, cfg['docs_dir'], cfg['site_dir'], cfg['use_directory_urls']) for f in files]
         pg = Page('Foo', fs[0], cfg)
         pg.read_source(cfg)
         pg.render(cfg, Files(fs))
@@ -706,139 +652,155 @@ class RelativePathExtensionTests(unittest.TestCase):
     def test_relative_html_link(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'non-index.md']),
-            '<p><a href="non-index/">link</a></p>'  # No trailing /
+            '<p><a href="non-index/">link</a></p>',  # No trailing /
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](index.md)'))
     def test_relative_html_link_index(self):
         self.assertEqual(
             self.get_rendered_result(['non-index.md', 'index.md']),
-            '<p><a href="..">link</a></p>'
+            '<p><a href="..">link</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/index.md)'))
     def test_relative_html_link_sub_index(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'sub2/index.md']),
-            '<p><a href="sub2/">link</a></p>'  # No trailing /
+            '<p><a href="sub2/">link</a></p>',  # No trailing /
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/non-index.md)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/non-index.md)')
+    )
     def test_relative_html_link_sub_page(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'sub2/non-index.md']),
-            '<p><a href="sub2/non-index/">link</a></p>'  # No trailing /
+            '<p><a href="sub2/non-index/">link</a></p>',  # No trailing /
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](file%20name.md)'))
     def test_relative_html_link_with_encoded_space(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'file name.md']),
-            '<p><a href="file%20name/">link</a></p>'
+            '<p><a href="file%20name/">link</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](file name.md)'))
     def test_relative_html_link_with_unencoded_space(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'file name.md']),
-            '<p><a href="file%20name/">link</a></p>'
+            '<p><a href="file%20name/">link</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](../index.md)'))
     def test_relative_html_link_parent_index(self):
         self.assertEqual(
             self.get_rendered_result(['sub2/non-index.md', 'index.md']),
-            '<p><a href="../..">link</a></p>'
+            '<p><a href="../..">link</a></p>',
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](non-index.md#hash)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open', mock.mock_open(read_data='[link](non-index.md#hash)')
+    )
     def test_relative_html_link_hash(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'non-index.md']),
-            '<p><a href="non-index/#hash">link</a></p>'
+            '<p><a href="non-index/#hash">link</a></p>',
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/index.md#hash)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/index.md#hash)')
+    )
     def test_relative_html_link_sub_index_hash(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'sub2/index.md']),
-            '<p><a href="sub2/#hash">link</a></p>'
+            '<p><a href="sub2/#hash">link</a></p>',
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/non-index.md#hash)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open', mock.mock_open(read_data='[link](sub2/non-index.md#hash)')
+    )
     def test_relative_html_link_sub_page_hash(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'sub2/non-index.md']),
-            '<p><a href="sub2/non-index/#hash">link</a></p>'
+            '<p><a href="sub2/non-index/#hash">link</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](#hash)'))
     def test_relative_html_link_hash_only(self):
         self.assertEqual(
             self.get_rendered_result(['index.md']),
-            '<p><a href="#hash">link</a></p>'
+            '<p><a href="#hash">link</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='![image](image.png)'))
     def test_relative_image_link_from_homepage(self):
         self.assertEqual(
             self.get_rendered_result(['index.md', 'image.png']),
-            '<p><img alt="image" src="image.png" /></p>'  # no opening ./
+            '<p><img alt="image" src="image.png" /></p>',  # no opening ./
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='![image](../image.png)'))
     def test_relative_image_link_from_subpage(self):
         self.assertEqual(
             self.get_rendered_result(['sub2/non-index.md', 'image.png']),
-            '<p><img alt="image" src="../../image.png" /></p>'
+            '<p><img alt="image" src="../../image.png" /></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='![image](image.png)'))
     def test_relative_image_link_from_sibling(self):
         self.assertEqual(
             self.get_rendered_result(['non-index.md', 'image.png']),
-            '<p><img alt="image" src="../image.png" /></p>'
+            '<p><img alt="image" src="../image.png" /></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='*__not__ a link*.'))
     def test_no_links(self):
         self.assertEqual(
             self.get_rendered_result(['index.md']),
-            '<p><em><strong>not</strong> a link</em>.</p>'
+            '<p><em><strong>not</strong> a link</em>.</p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[link](non-existent.md)'))
     def test_bad_relative_html_link(self):
-        with self.assertLogs('mkdocs', level='WARNING') as cm:
+        with self.assertLogs('mkdocs') as cm:
             self.assertEqual(
                 self.get_rendered_result(['index.md']),
-                '<p><a href="non-existent.md">link</a></p>'
+                '<p><a href="non-existent.md">link</a></p>',
             )
         self.assertEqual(
-            cm.output,
-            ["WARNING:mkdocs.structure.pages:Documentation file 'index.md' contains a link "
-             "to 'non-existent.md' which is not found in the documentation files."]
+            '\n'.join(cm.output),
+            "WARNING:mkdocs.structure.pages:Documentation file 'index.md' contains a link "
+            "to 'non-existent.md' which is not found in the documentation files.",
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[external](http://example.com/index.md)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open',
+        mock.mock_open(read_data='[external](http://example.com/index.md)'),
+    )
     def test_external_link(self):
         self.assertEqual(
             self.get_rendered_result(['index.md']),
-            '<p><a href="http://example.com/index.md">external</a></p>'
+            '<p><a href="http://example.com/index.md">external</a></p>',
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[absolute link](/path/to/file.md)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open', mock.mock_open(read_data='[absolute link](/path/to/file.md)')
+    )
     def test_absolute_link(self):
         self.assertEqual(
             self.get_rendered_result(['index.md']),
-            '<p><a href="/path/to/file.md">absolute link</a></p>'
+            '<p><a href="/path/to/file.md">absolute link</a></p>',
         )
 
-    @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='[absolute local path](\\image.png)'))
+    @mock.patch(
+        'mkdocs.structure.pages.open',
+        mock.mock_open(read_data='[absolute local path](\\image.png)'),
+    )
     def test_absolute_win_local_path(self):
         self.assertEqual(
             self.get_rendered_result(['index.md']),
-            '<p><a href="\\image.png">absolute local path</a></p>'
+            '<p><a href="\\image.png">absolute local path</a></p>',
         )
 
     @mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data='<mail@example.com>'))
@@ -849,5 +811,5 @@ class RelativePathExtensionTests(unittest.TestCase):
             # The following is equivalent to: '<p><a href="mailto:mail@example.com">mail@example.com</a></p>'
             '<p><a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#109;&#97;&#105;&#108;&#64;&#101;'
             '&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;">&#109;&#97;&#105;&#108;&#64;'
-            '&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;</a></p>'
+            '&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;</a></p>',
         )
